/*
AddAppInfo, Copyright (C) 1999-2000 Jean-Baptiste M. Queru

This program is free software; you can redistribute it and/or 
modify it under the terms of the enclosed JBQ code license.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
License for more details. 

Contact the author at:
Jean-Baptiste M. Queru, 31 Camino Alto, Millbrae CA 94030, USA
or by e-mail at : jb@queru.com
*/

#include <Application.h>
#include <AppFileInfo.h>
#include <View.h>
#include <Bitmap.h>
#include <TranslationUtils.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <Path.h>
#include <Entry.h>
#include <Roster.h>

#include "AddAppInfo.h"
#include "ParseHelpers.h"
#include "Macros.h"

int mainret=0;

keyval main_keywords[]={
	{"app_flags",0},
	{"app_version",1},
	{"system_version",2},
	{"signature",3},
	{"supported_types",4},
	{"supported_types_nosync",5},
	{"supported_types_sync",6},
	{"mini_icon",7},
	{"large_icon",8},
	{"mini_icon_type",9},
	{"large_icon_type",10},
	{NULL,0xDEADBEEF}
};

keyval app_flags_values[]={
	NAMEVAL(B_SINGLE_LAUNCH)
	NAMEVAL(B_MULTIPLE_LAUNCH)
	NAMEVAL(B_EXCLUSIVE_LAUNCH)
	NAMEVAL(B_BACKGROUND_APP)
	NAMEVAL(B_ARGV_ONLY)
	{NULL,0xDEADBEEF}
};

AAIApp::AAIApp():BApplication("application/x-vnd.djaybee.addappinfo") {
	srcname=dstname=NULL;
	earlyquit=false;
}

// Parse the arguments as they are received, to be handled in ReadyToRun
// XXX jbq should check for IsLaunching
void AAIApp::ArgvReceived(int32 argc,char** argv) {
// if any argument is --help or --format, handle if first and bail out
// XXX jbq : that makes if impossible to have a file called with those options names
	for (int32 i=1;i<argc;i++) {
		if ((strcmp(argv[i],"--help")==0)||(strcmp(argv[i],"-h")==0)) {
			printf("usage : %s [{--input,-i}] <srcfile> [{--output,-o}] <dstfile>\n",argv[0]);
			printf("Parses the content of srcfile and writes it into dstfile's app_info\n\n");
			printf("use '/' as the srcfile to read input from stdin\n\n");
			printf("%s --help   [-h] : prints this help to stdout\n",argv[0]);
			printf("%s --format [-f] : prints the input file format\n",argv[0]);
			earlyquit=true;
			OKBYE;			
		} else if ((strcmp(argv[i],"--format")==0)||(strcmp(argv[i],"-f")==0)) {
			printf("%s input file format : \n\n",argv[0]);
			printf("Each line contains a keyword and its parameters :\n");
			printf("app_flags <flag> [<flag> [...]]\n");
			printf("{app,system}_version <major> <middle> <minor> <variety> <internal> ...\n");
			printf("                                                    ...<short info> <long info>\n");
			printf("signature <mime string>\n");
			printf("supported_types{,_sync,_nosync} <mime string> [<mime string> [...]]\n");
			printf("{mini,large}_icon <filename> [transparent_color]\n");
			printf("{mini,large}_icon_type <filename> <mime string> [transparent_color]\n\n");
			printf("Special characters :\n");
			printf("# begins a comment\n");
			printf("\\ can be used to escape special characters inside words\n");
			earlyquit=true;
			OKBYE;
		}
	}
	for (int32 i=1;i<argc;i++) {
		if (((strcmp(argv[i],"-i")==0)||(strcmp(argv[i],"--input")==0))&&(i!=argc-1)&&(srcname==NULL)) {
			srcname=strdup(argv[++i]);
		} else if (((strcmp(argv[i],"-o")==0)||(strcmp(argv[i],"--output")==0))&&(i!=argc-1)&&(dstname==NULL)) {
			dstname=strdup(argv[++i]);
		} else if (srcname==NULL) {
			srcname=strdup(argv[i]);
		} else if (dstname==NULL) {
			dstname=strdup(argv[i]);
		} else {
			PrintUsage(argv[0]);
			earlyquit=true;
			BYE;
		}
	}
}

void AAIApp::PrintUsage(const char*n) {
	if (n==NULL) {
		app_info ai;
		BPath p;
		GetAppInfo(&ai);
		BEntry(&ai.ref).GetPath(&p);
		n=strdup(p.Path());
	}
	fprintf(stderr,"usage : %s <srcfile> <dstfile>\n",n);
	fprintf(stderr,"        %s --help\n",n);
}

// where the actual work happens
void AAIApp::ReadyToRun() {
	if (earlyquit) return; // don't do anything if asked so by earlier functions
	int srcline=1;
	off_t srcsize;
	char* srcdata;

// flags that are used to guarantee that a given information isn't written twice
	bool has_app_flags=false,has_app_version=false,has_system_version=false,has_signature=false;
	bool has_supported_types=false,has_mini_icon=false,has_large_icon=false;

// must also check here in case it was called with no arguments
	if ((srcname==NULL)||(dstname==NULL)) {
		PrintUsage();
		BYE;
	}

	if (strcmp(srcname,"/")==0) { // read data from stdin, by chunks of 4kB at a time. always keep an extra byte around for NULL-termination.
		srcdata=(char*)malloc(4097);
		int sread=0,scurrent;
		while((scurrent=fread(srcdata+sread,1,4096,stdin))!=0) {
			sread+=scurrent;
			srcdata=(char*)realloc(srcdata,sread+4097);
		}
		if (!feof(stdin)) {
			SOMEERROR("could not read input from stdin");
		}
		srcdata[sread]='\0';
	} else { // open the file, get its size, allocate space for data, read the file
		BFile src(srcname,B_READ_ONLY);
		if (src.InitCheck()!=B_OK) {
			NAMEERROR("could not open source file",srcname);
		}
		src.GetSize(&srcsize);
		srcdata=(char*)malloc(srcsize+1);
		if (srcdata==NULL) {
			NAMEERROR("could not allocate space to read source file",srcname);
		}
// XXX jbq - not needed. only clear the last byte, and do it after reading
		memset(srcdata,0,srcsize+1);
// XXX jbq - should try to read in a loop, in case it gets interrpted. should try to read an extra byte past the end, just in case
		if (src.Read(srcdata,srcsize)!=srcsize) {
			NAMEERROR("error while reading source file",srcname);
		}
	}

// create the app info. an invalid BFile will create an invalid BAppFileInfo, so don't check for BFile::InitCheck
	BAppFileInfo dst(new BFile(dstname,B_READ_WRITE));
	if (dst.InitCheck()!=B_OK) {
		NAMEERROR("could not open destination file",dstname);
	}
	dst.SetInfoLocation(B_USE_BOTH_LOCATIONS);

// the main loop
	while (true) {
		srcline+=skiplinecount(&srcdata); // go to the next word

		if (*srcdata=='\0') { // end of file - this is where the program normally exits
			break;
		}

// command dispatch
		int32 kw;
		if (findvalue(srcdata,main_keywords,&kw)) {
			switch(kw) {
				case 0 : { // app_flags
					if (has_app_flags) {
						LINEERROR("cannot set the app_flags twice");
					}
					has_app_flags=true;
					bool found=false;
					int32 afval=0;
					NEXTWORD;
					while (!endofline(srcdata)) {
						if (findorvalue(srcdata,app_flags_values,&afval)) {
							found=true;
							NEXTWORD;
						} else {
							PARSEERROR("unknown app_flags parameter");
						}
					}
					if (!found) {
						LINEERROR("app_flags needs some parameters");
					}
					if (dst.SetAppFlags(afval)!=B_OK) {
						LINEERROR("error writing the app_flags");
					}
					break;
				}
				case 1 : // app version
				case 2 : { // system version
					if (kw==1) {
						if (has_app_version) {
							LINEERROR("cannot set the app_version twice");
						}
						has_app_version=true;
					} else {
						if (has_system_version) {
							LINEERROR("cannot set the system_version twice");
						}
						has_system_version=true;
					}
					NEXTWORD;
					version_info vi;
					for (int i=0;i<5;i++) {
						uint32* vpart=NULL;
						switch(i) {
							case 0: vpart=&vi.major;break;
							case 1: vpart=&vi.middle;break;
							case 2: vpart=&vi.minor;break;
							case 3: vpart=&vi.variety;break;
							case 4: vpart=&vi.internal;break;
						}
						if (extractnumber(srcdata,vpart)) {
							NEXTWORD;
						} else {
							PARSEERROR("number expected for version info");
						}
					}
					char*svers=extractword(&srcdata);
					if (svers==NULL) {
						LINEERROR("expected short version info");
					}
					char*lvers=extractword(&srcdata);
					if (lvers==NULL) {
						LINEERROR("expected long version info");
					}
					strncpy(vi.short_info,svers,64);
					strncpy(vi.long_info,lvers,256);
					if (dst.SetVersionInfo(&vi,(kw==1)?B_APP_VERSION_KIND:B_SYSTEM_VERSION_KIND)!=B_OK) {
						LINEERROR("error writing the version_info");
					}
					free(svers);
					free(lvers);
					break;
				}
				case 3 : { // signature
					if (has_signature) {
						LINEERROR("cannot set the signature twice");
					}
					has_signature=true;
					NEXTWORD;
					char*sig=extractword(&srcdata);
					if (sig==NULL) {
						LINEERROR("expected a signature");
					}
					if (dst.SetSignature(sig)!=B_OK) {
						LINEERROR("error writing the signature");
					}
					free(sig);
					break;
				}
				case 4 : // supported_types
				case 5 : // supported_types_nosync
				case 6 : { // supported_types_sync
					if (has_supported_types) {
						LINEERROR("cannot set the supported_types twice");
					}
					has_supported_types=true;
					NEXTWORD;
					BMessage types;
					char*type;
					while ((type=extractword(&srcdata))!=NULL) {
						types.AddString("types",type);
						free(type);
					}
					if (!types.HasString("types")) {
						LINEERROR("supported_types needs a list of types");
					}
					status_t ret=B_ERROR;
					switch(kw) {
						case 4 : ret=dst.SetSupportedTypes(&types);break;
						case 5 : ret=dst.SetSupportedTypes(&types,false);break;
						case 6 : ret=dst.SetSupportedTypes(&types,true);break;
					}
					if (ret!=B_OK) {
						LINEERROR("error writing the supported_types");
					}
					break;
				}
				case 7 : // mini_icon
				case 8 : { // large_icon
					if (kw==7) {
						if (has_mini_icon) {
							LINEERROR("cannot set the mini_icon twice");
						}
						has_mini_icon=true;
					} else {
						if (has_large_icon) {
							LINEERROR("cannot set the large_icon twice");
						}
						has_large_icon=true;
					}
					NEXTWORD;
					char*name=extractword(&srcdata);
					if (name==NULL) {
						LINEERROR("icon needs a filename");
					}
					BBitmap*origbmp=BTranslationUtils::GetBitmapFile(name);
					if (origbmp==NULL) {
						LINENAMEERROR("icon could not load bitmap",name);
					}
					free(name);
					BBitmap*bmp=new BBitmap(BRect(0,0,(kw==7)?15:31,(kw==7)?15:31),B_CMAP8,true);
					BView*v=new BView(BRect(0,0,31,31),"",0,0);
					bmp->Lock();
					bmp->AddChild(v);
					memset(bmp->Bits(),B_TRANSPARENT_8_BIT,bmp->BitsLength());
					v->DrawBitmap(origbmp,BPoint(0,0));
					if (!endofline(srcdata)) {
						char*transstr=extractword(&srcdata);
						uint32 trans=strtoul(transstr,NULL,0);
						free(transstr);
						if (origbmp->BitsLength()<4*origbmp->BytesPerRow()*(origbmp->Bounds().bottom+1)) {
							int xmax=(kw==7)?15:31;
							if (xmax>origbmp->Bounds().right) xmax=(int)origbmp->Bounds().right;
							int ymax=(kw==7)?15:31;
							if (ymax>origbmp->Bounds().bottom) ymax=(int)origbmp->Bounds().bottom;
							for (int y=0;y<=ymax;y++) {
								for (int x=0;x<=xmax;x++) {
									if (((uint32*)(origbmp->Bits()))[x+y*origbmp->BytesPerRow()/4]==trans) {
										((char*)(bmp->Bits()))[x+y*bmp->BytesPerRow()]=B_TRANSPARENT_8_BIT;
									}
								}
							}
						}
					}
					if (dst.SetIcon(bmp,(kw==7)?B_MINI_ICON:B_LARGE_ICON)!=B_OK) {
						LINEERROR("error setting the icon");
					}
					bmp->Unlock();
					delete origbmp;
					delete bmp;
					break;
				}
				case 9 : // mini_icon_type
				case 10 : { // large_icon_type
					NEXTWORD;
					char*name=extractword(&srcdata);
					if (name==NULL) {
						LINEERROR("icon_type needs a filename");
					}
					char*sig=extractword(&srcdata);
					if (sig==NULL) {
						LINEERROR("icon_type needs a type");
					}
					BBitmap*origbmp=BTranslationUtils::GetBitmapFile(name);
					if (origbmp==NULL) {
						LINENAMEERROR("icon could not load bitmap",name);
					}
					free(name);
					BBitmap*bmp=new BBitmap(BRect(0,0,(kw==9)?15:31,(kw==9)?15:31),B_CMAP8,true);
					BView*v=new BView(BRect(0,0,31,31),"",0,0);
					bmp->Lock();
					bmp->AddChild(v);
					v->SetDrawingMode(B_OP_OVER);
					memset(bmp->Bits(),B_TRANSPARENT_8_BIT,bmp->BitsLength());
					v->DrawBitmap(origbmp,BPoint(0,0));
					if (!endofline(srcdata)) {
						char*transstr=extractword(&srcdata);
						uint32 trans=strtoul(transstr,NULL,0);
						free(transstr);
						if (origbmp->BitsLength()<4*origbmp->BytesPerRow()*(origbmp->Bounds().bottom+1)) {
							int xmax=(kw==7)?15:31;
							if (xmax>origbmp->Bounds().right) xmax=(int)origbmp->Bounds().right;
							int ymax=(kw==7)?15:31;
							if (ymax>origbmp->Bounds().bottom) ymax=(int)origbmp->Bounds().bottom;
							for (int y=0;y<=ymax;y++) {
								for (int x=0;x<=xmax;x++) {
									if (((uint32*)(origbmp->Bits()))[x+y*origbmp->BytesPerRow()/4]==trans) {
										((char*)(bmp->Bits()))[x+y*bmp->BytesPerRow()]=B_TRANSPARENT_8_BIT;
									}
								}
							}
						}
					}
					if (dst.SetIconForType(sig,bmp,(kw==9)?B_MINI_ICON:B_LARGE_ICON)!=B_OK) {
						LINENAMEERROR("error setting the icon for type",sig);
					}
					bmp->Unlock();
					delete origbmp;
					delete bmp;
					free(sig);
					break;
				}
			}
		} else {
			PARSEERROR("unknown command");
		}
		if (!endofline(srcdata)) { // this happens when a command reads all its parameters but some others are found on the same line
			PARSEERROR("unexpected characters at end of line");
		}
	}
	PostMessage(B_QUIT_REQUESTED);
}

// everything happens in ArgvReceived. note that the app *must* be multiple-launch
int main() {
	AAIApp().Run();
	return mainret;
}
