/*
AddAppInfo, Copyright (C) 1999-2000 Jean-Baptiste M. Queru

This program is free software; you can redistribute it and/or
modify it under the terms of the enclosed JBQ code license.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
License for more details.

Contact the author at:
Jean-Baptiste M. Queru, 31 Camino Alto, Millbrae CA 94030, USA
or by e-mail at : jb@queru.com
*/

#include "ParseHelpers.h"

#include <string.h>
#include <stdlib.h>

// a bunch of helper functions

// return the beginning of the next word
// XXX jbq : I'm not sure if this actually works.
char* nextword(char*t) {
	// move until reaching the end of the current word, if needed.
	// XXX jbq : what about using endofword() ?
	while ((*t!='\r')&&(*t!='\n')&&(*t!='\t')&&(*t!='#')&&(*t!=' ')&&(*t!='\0')) t++;
	// move until the beginning of the next word.
	while ((*t==' ')||(*t=='\t')) t++;
	return t;
}

// insert a 0 at the end of the current word
void cutword(char*t) {
	t=strpbrk(t,"\n\r\t# ");
	if (t!=NULL) *t='\0';
}

// tests is pointer points to an end of line
bool endofline(char*t) {
	return (*t=='\r')||(*t=='\n')||(*t=='#')||(*t=='\0');
}

// tests is pointer points to an end of word (i.e. any char that doesn't belong to a word)
bool endofword(char*t) {
	return (*t=='\r')||(*t=='\n')||(*t=='#')||(*t=='\0')||(*t==' ')||(*t=='\t');
}

// finds a value in a key/value array. returns true if found.
bool findvalue(char*t,keyval*vals,int32*val) {
	while(vals->name!=NULL) {
		// only compare for the length of the keyword, and check that the parsed word isn't longer than the keyword.
		// XXX jbq : potentially unexpected results if the keyword contains some "endofword" characters
		if ((strncmp(t,vals->name,strlen(vals->name))==0)&&(endofword(t+strlen(vals->name)))) {
			*val=vals->value;
			return true;
		}
		vals++;
	}
	return false;
}

// finds a value in a key/value array, and or's it with the current value. returns true if found.
bool findorvalue(char*t,keyval*vals,int32*val) {
	int32 tv;
	if (findvalue(t,vals,&tv)) {
		(*val)|=tv;
		return true;
	} else {
		return false;
	}
}

// skip characters until something interesting happens. returns the number of lines that were skipped.
int skiplinecount(char**t) {
	int l=0;
	while (true) {
		if (**t=='\r') { // skip a carriage return and increment the linecount.
			l++;
			(*t)++;
			if (**t=='\n') { // if the carriage return is followed by a newline, skip the newline without counting it
				(*t)++;
			}
		} else if (**t=='\n') { // skip a newline and incement the linecount
			l++;
			(*t)++;
		} else if ((**t==' ')||(**t=='\t')) { // skip spaces too
			(*t)++;
		} else if (**t=='#') { // skip the beginning of a comment
			while ((**t!='\r')&&(**t!='\n')&&(**t!='\0')) { // until end of line/file
				(*t)++;
			}
		} else {
			break;
		}
	}
	return l; // this is the letter 'ell', not the number 'one'
}

// extracts an escaped word from the stream and advances the pointer accordingly
// XXX jbq - could be optimized. who cares?
char* extractword(char**t) {
	int l=1; // length of the current string, including the null-termination
	char*ret=strdup(""); // first initialization - XXX jbq check return value
	while(!endofword(*t)) { // loop until reaching end of word
		ret=(char*)realloc(ret,l+1); // allocate space for one more byte - XXX jbq check return value
		ret[l]='\0'; // null-terminate for safety
		char nchar=*((*t)++); // read the current char. guaranteed not to be an end-of-word
		if (nchar!='\\') { // not an escape char
			ret[l++-1]=nchar; // copy over
		} else {
			if (**t=='\0') { // ignore if an escape is the last character of the file
				break;
			} else if (**t=='\r') { // special-case \CR and \CRLF, replace with LF. \LF is covered by the default case
				(*t)++;
				if (**t=='\n') {
					(*t)++;
				}
				ret[l++-1]='\n';
			} else {
				ret[l++-1]=*((*t)++); // no special-case, copy the next byte over.
			}
		}
	}
	*t=nextword(*t); // reached the end of the word. jump to the next one
	if (strlen(ret)==0) { // if no byte was found, don't return anything
		delete ret;
		return NULL;
	} else {
		return ret;
	}
}

// extracts a number from the stream. return true if found.
bool extractnumber(char*t,uint32*n) {
	*n=0; // initialize with 0
	bool found=false;
	while((*t>='0')&&(*t<='9')) { // move forward as long as digits are found
		*n*=10;
		*n+=(*(t++)-'0');
		found=true;
	}
	return found&&endofword(t); // fail if the number wasn't a whole word, or if no digit was found
}

